# Open and analyze UCN data for a one cycle
# Derek Fujimoto
# Oct 2024

from .exceptions import *
from .ucnbase import ucnbase
from .tsubfile import tsubfile
from .ttreeslow import ttreeslow

import numpy as np
import os

class ucnperiod(ucnbase):
    """Stores the data from a single UCN period from a single cycle

    Args:
        ucycle (ucncycle): object to pull period from
        period (int): period number
    """

    def __init__(self, ucycle, period):

        # get start and stop time
        if period > 0:      start = ucycle.cycle_param.period_end_times[period-1]
        else:               start = ucycle.cycle_start

        stop = ucycle.cycle_param.period_end_times[period]

        # copy data
        for key, value in ucycle.__dict__.items():
            if key == 'tfile':
                setattr(self, key, tsubfile(value, start, stop))
            elif key == 'epics':
                setattr(self, key, ttreeslow(value, self))
            elif hasattr(value, 'copy'):
                setattr(self, key, value.copy())
            else:
                setattr(self, key, value)

        # trim cycle parameters
        self.cycle_param.period_durations_s = self.cycle_param.period_durations_s[period]
        self.cycle_param.period_end_times = self.cycle_param.period_end_times[period]

        self.period = period
        self.period_start = start
        self.period_stop = stop
        self.period_dur = stop-start

    def __repr__(self):
        klist = [d for d in self.__dict__.keys() if d[0] != '_']
        if klist:

            # sort without caps
            klist.sort(key=lambda x: x.lower())

            # get number of columns based on terminal size
            maxsize = max((len(k) for k in klist)) + 2
            terminal_width = os.get_terminal_size().columns
            ncolumns = int(np.floor(terminal_width / maxsize))
            ncolumns = min(ncolumns, len(klist))

            # split into chunks
            needed_len = int(np.ceil(len(klist) / ncolumns)*ncolumns) - len(klist)
            klist = np.concatenate((klist, np.full(needed_len, '')))
            klist = np.array_split(klist, ncolumns)

            # print
            cyc_str = f' (cycle {self.cycle}, period {self.period})'
            s = f'run {self.run_number}{cyc_str}:\n'
            for key in zip(*klist):
                s += '  '
                s += ''.join(['{0: <{1}}'.format(k, maxsize) for k in key])
                s += '\n'
            return s
        else:
            return self.__class__.__name__ + "()"

    def is_pileup(self, detector):
        """Check if pileup may be an issue in this period.

        Histograms the first `pileup_within_first_s` seconds of data in 1 ms bins and checks if any of those bins are greater than the `pileup_cnt_per_ms` threshold.

        Args:
            detector (str): one of the keys to self.DET_NAMES

        Returns:
            bool: true if pileup detected

        Example:
            ```python
            >>> p = run[0, 0]
            >>> p.is_pileup('Li6')
            False
            ```
        """

        # get the tree
        hit_tree = super().get_hits(detector)

        ## filter pileup for period data

        # get thresholds
        dt = self.DATA_CHECK_THRESH['pileup_within_first_s']
        count_thresh = self.DATA_CHECK_THRESH['pileup_cnt_per_ms']

        # make histogram
        t = hit_tree.index.values
        counts, _ = np.histogram(t, bins=int(1/0.001*dt),
                                        range=(min(t), min(t)+dt))

        # look for pileup
        piled_up = counts > count_thresh

        return any(piled_up)

    def get_nhits(self, detector):
        """Get number of ucn hits

        Args:
            detector (str): Li6|He3
        """
        return self._run._get_nhits(detector, cycle=self.cycle, period=self.period)

    def modify_timing(self, dt_start_s=0, dt_stop_s=0):
        """Change start and end times of period

        Args:
            dt_start_s (float): change to the period start time
            dt_stop_s (float): change to the period stop time

        Notes:
            * as a result of this, cycles may overlap or have gaps
            * periods are forced to not overlap and have no gaps
            * cannot change cycle end time, but can change cycle start time
            * this function resets all saved histgrams and hits
        """
        self._run._modify_ptiming(cycle = self.cycle,
                                  period = self.period,
                                  dt_start_s = dt_start_s,
                                  dt_stop_s = dt_stop_s)


# Fetch a subrange of tfile object
# Derek Fujimoto
# Oct 2024

from .exceptions import *
import numpy as np
import pandas as pd
from rootloader import tfile, ttree

class tsubfile(tfile):
    """Wrapper for tfile which restricts access to values only within given times

    Args:
        tfileobj (tfile): object to wrap
        start (int): starting epoch time
        stop (int): stopping epoch time
    """

    def __init__(self, tfileobj, start, stop):

        for key, value in tfileobj.items():
            self[key] = value

        items = {'_start':start, '_stop':stop}
        object.__setattr__(self, '_items', items)

    def __getitem__(self, key):

        # quick return
        if key in self._items.keys():
            return self._items[key]

        # get the data
        val = super().__getitem__(key)

        # get sub range: pd.DataFrame
        if isinstance(val, pd.DataFrame):
            try:
                index_name = val.index.name.lower()
            except AttributeError:
                pass
            else:
                if 'time' in index_name:
                    idx = (val.index >= self._start) & (val.index <= self._stop)
                    val = val.loc[idx]

        # get sub range: rootloader.ttree
        elif isinstance(val, ttree):
            try:
                index_name = val.index_name.lower()
            except AttributeError:
                pass
            else:
                if 'time' in index_name:
                    val = val.loc[self._start:self._stop]

        self._items[key] = val

        return val

    def __getattr__(self, name):

        if name in self.keys():
            return self[name]
        elif name in self._items.keys():
            return self._items[name]
        else:
            return self.__getattribute__(name)




# Array which syntatic sugar to apply functions to each element contained
# Derek Fujimoto
# Oct 2024

import numpy as np

class applylist(list):
    """A list object with the following enhancements:

    * An apply function: like in pandas, apply takes a function handle and applies it to every element in the list. This acts recursively, if the list has a depth of more than 1
    * Element access: accessing attributes, if not an attribute of the applylist, instead try to fetch attributes of the contained objects. The same for functions. This also works recursively.
    * Numpy-like array slicing: slicing with a np.ndarray first converts this object to an array, then does the slice, then converts back. This allows slicing on arrays of indices for re-ordering or booleans for selection based on criteria
    * Arithmetic works element-wise as in numpy arrays

    Examples:

        >>> # Slicing
        >>> x = applylist(range(10))
        >>> print(x[3:7])
        [3, 4, 5, 6]

        >>> # Element access
        >>> x = applylist([ucnrun(1846), ucnrun(1847)])
        >>> print(x.run_number)
        [1846, 1847]
        >>> print(x.beam_current_uA.mean())
        [np.float64(0.16612837637441483), np.float64(0.18927602913972205)]

        >>> # Arithmetic and comparisons
        >>> x = applylist([1,2,3])
        >>> print(x*2)
        [np.int64(2), np.int64(4), np.int64(6)]
        >>> print(x>2)
        [np.False_, np.False_, np.True_]
    """

    def __call__(self, *args, **kwargs):

        # call function of sub-elements
        return applylist([x.__call__(*args, **kwargs) for x in self])

    # numpy-style comparisons
    def __eq__(self, other):    return applylist(np.array(self).__eq__(other))
    def __ne__(self, other):    return applylist(np.array(self).__ne__(other))
    def __lt__(self, other):    return applylist(np.array(self).__lt__(other))
    def __gt__(self, other):    return applylist(np.array(self).__gt__(other))
    def __le__(self, other):    return applylist(np.array(self).__le__(other))
    def __ge__(self, other):    return applylist(np.array(self).__ge__(other))

    # numpy-style unary operators
    def __pos__(self):          return applylist(np.array(self).__pos__())
    def __neg__(self):          return applylist(np.array(self).__neg__())
    def __abs__(self):          return applylist(np.array(self).__abs__())
    def __invert__(self):       return applylist(np.array(self).__invert__())
    def __round__(self):        return applylist(np.array(self).__round__())
    def __floor__(self):        return applylist(np.array(self).__floor__())
    def __ceil__(self):         return applylist(np.array(self).__ceil__())
    def __trunc__(self):        return applylist(np.array(self).__trunc__())

    # numpy-style arithmetic operators
    def __add__(self, other):   return applylist(np.array(self).__add__(other))
    def __sub__(self, other):   return applylist(np.array(self).__sub__(other))
    def __mul__(self, other):   return applylist(np.array(self).__mul__(other))
    def __floordiv__(self, other):   return applylist(np.array(self).__floordiv__(other))
    def __div__(self, other):   return applylist(np.array(self).__div__(other))
    def __truediv__(self, other):   return applylist(np.array(self).__truediv__(other))
    def __mod__(self, other):   return applylist(np.array(self).__mod__(other))
    def __divmod__(self, other):   return applylist(np.array(self).__divmod__(other))
    def __pow__(self, other):   return applylist(np.array(self).__pow__(other))
    def __and__(self, other):   return applylist(np.array(self).__and__(other))
    def __or__(self, other):   return applylist(np.array(self).__or__(other))
    def __xor__(self, other):   return applylist(np.array(self).__xor__(other))

    def __getattr__(self, name):

        # retain default behaviour
        try:
            return super().__getattr__(name)

        # enhanced behaviour
        except AttributeError as err:

            # check for scalar numpy datatypes, all contents assumed to be of the same type
            if isinstance(self[0], (int, float, np.integer, np.floating)) or '__array_' in name:
                raise err from None

            # try to get the contents of subarray
            return applylist([getattr(x, name) for x in self])

    def __getitem__(self, key):

        # numpy-like slicing
        if isinstance(key, (np.ndarray, tuple, list)):
            return applylist(np.array(self)[key])

        # default behaviour
        else:
            return super().__getitem__(key)

    def astype(self, typecast):
        """Convert datatypes in self to typecast

        Args:
            typecase (type): type to convert to

        Returns:
            None, works in-place

        Example:

            >>> x = applylist(np.arange(5))
            >>> print(x)
            [np.int64(0), np.int64(1), np.int64(2), np.int64(3), np.int64(4)]
            >>> x.astype(int)
            >>> print(x)
            [0, 1, 2, 3, 4]
        """
        for i in range(len(self)):
            self[i] = typecast(self[i])

    def apply(self, fn, inplace=False):
        """Apply function to each element contained, similar to pandas functionality

        Args:
            fn (function handle): function to apply to each element
            inplace (bool): if false return a copy, else act in-place

        Returns:
            ucnarray|None: depending on the value of inplace

        Examples:

            >>> # With return value
            >>> x = applylist(np.arange(5))
            >>> y = x.apply(lambda a: a**2)
            >>> print(x)
            [np.int64(0), np.int64(1), np.int64(2), np.int64(3), np.int64(4)]
            >>> print(y)
            [np.int64(0), np.int64(1), np.int64(4), np.int64(9), np.int64(16)]

            >>> # Inplace
            >>> x = applylist(np.arange(5))
            >>> x.apply(lambda a: a**2, inplace=True)
            >>> print(x)
            [np.int64(0), np.int64(1), np.int64(4), np.int64(9), np.int64(16)]
        """

        # decide if inplace
        if inplace: copy = self
        else:       copy = self.copy()

        # do the operation
        for i in range(len(copy)):
            if isinstance(copy[i], applylist):
                copy[i] = copy[i].apply(fn)
            else:
                copy[i] = fn(copy[i])

        # return the copy
        if not inplace: return copy

    def transpose(self):
        """Transpose by conversion to np.array and back

        Args:
            None

        Returns:
            applylist: transposed

        Example:

            >>> x = applylist([[1,2,3], [4,5,6]])
            >>> print(x)
            [[1, 2, 3], [4, 5, 6]]
            >>> print(x.transpose())
            [array([1, 4]), array([2, 5]), array([3, 6])]
        """
        return applylist(np.array(self).transpose())



# Open and analyze UCN data for a one cycle
# Derek Fujimoto
# Oct 2024
from .exceptions import *
from .applylist import applylist
from .ucnbase import ucnbase
from .ucnperiod import ucnperiod
from .tsubfile import tsubfile
from .ttreeslow import ttreeslow
from tqdm import tqdm


import numpy as np
import os

class ucncycle(ucnbase):
    """View for the data from a single UCN cycle

    Notes:
        Any changes to the data frame affects all periods (for the time steps
        contained in that period) and the containing run

    Args:
        urun (ucnrun): object to pull cycle from
        cycle (int): cycle number
    """

    def __init__(self, urun, cycle):

        # get cycles to keep
        cycles = urun.cycle_param.cycle_times
        start = int(cycles.loc[cycle, 'start'])
        stop = int(cycles.loc[cycle, 'stop'])
        supercycle = int(cycles.loc[cycle, 'supercycle'])

        # copy data
        for key, value in urun.__dict__.items():
            if key == 'tfile':
                setattr(self, key, tsubfile(value, start, stop))
            elif key == 'epics':
                setattr(self, key, ttreeslow(value, self))
            elif hasattr(value, 'copy'):
                setattr(self, key, value.copy())
            else:
                setattr(self, key, value)

        # trim cycle parameters
        self.cycle_param.period_durations_s = self.cycle_param.period_durations_s[cycle]
        self.cycle_param.period_end_times = self.cycle_param.period_end_times[cycle]

        if self.cycle_param.filter is not None:
            self.cycle_param.filter = self.cycle_param.filter[cycle]

        self.cycle = cycle
        self.supercycle = supercycle
        self.cycle_start = start
        self.cycle_stop = stop
        self.cycle_dur = stop-start

        # store fetched periods
        self._perioddict = dict()

    def __len__(self):
        return self.cycle_param.nperiods

    def __next__(self):
        # permit iteration over object like it was a list

        # iterate
        if self._iter_current < self.cycle_param.nperiods:
            cyc = self[self._iter_current]
            self._iter_current += 1
            return cyc

        # end of iteration
        else:
            raise StopIteration()

    def __repr__(self):
        klist = [d for d in self.__dict__.keys() if d[0] != '_']
        if klist:

            # sort without caps
            klist.sort(key=lambda x: x.lower())

            # get number of columns based on terminal size
            maxsize = max((len(k) for k in klist)) + 2
            terminal_width = os.get_terminal_size().columns
            ncolumns = int(np.floor(terminal_width / maxsize))
            ncolumns = min(ncolumns, len(klist))

            # split into chunks
            needed_len = int(np.ceil(len(klist) / ncolumns)*ncolumns) - len(klist)
            klist = np.concatenate((klist, np.full(needed_len, '')))
            klist = np.array_split(klist, ncolumns)

            # print
            cyc_str = '' if self.cycle is None else f' (cycle {self.cycle})'
            s = f'run {self.run_number}{cyc_str}:\n'
            for key in zip(*klist):
                s += '  '
                s += ''.join(['{0: <{1}}'.format(k, maxsize) for k in key])
                s += '\n'
            return s
        else:
            return self.__class__.__name__ + "()"

    def __getitem__(self, key):
        # get cycle or period based on slicing indexes

        # get a single key
        if isinstance(key, (np.integer, int)):
            if key < 0:
                key = self.cycle_param.nperiods + key

            if key > self.cycle_param.nperiods:
                raise IndexError(f'Run {self.run_number}, cycle {self.cycle}: Index larger than number of periods ({self.cycle_param.nperiods})')

            return self.get_period(key)

        # slice on cycles
        if isinstance(key, slice):
            period = self.get_period()[:self.cycle_param.nperiods]
            return applylist(period[key])

        raise IndexError('Cycles indexable only as a 1-dimensional object')

    def check_data(self, raise_error=False, quiet=False):
        """Run some checks to determine if the data is ok.

        Args:
            period_production (int): index of period where the beam should be stable. Enables checks of beam stability
            period_count (int): index of period where we count ucn. Enables checks of data quantity
            period_background (int): index of period where we do not count ucn. Enables checks of background
            raise_error (bool): if true, raise an error if check fails, else return false. Inactive if quiet=True
            quiet (bool): if true don't print or raise exception

        Returns:
            bool: true if check passes, else false.

        Note:

            Checks performed

            * is there BeamlineEpics data?
            * is the cycle duration greater than 0?
            * is at least one valve opened during at least one period?
            * are there counts in each detector?

        Example:
            ```python
            >>> cycle = run[0]
            >>> x = cycle.check_data()
            Run 1846, cycle 0: Beam current dropped to 0.0 uA
            >>> x
            False
            ```
        """
        # setup error message
        msg = f'Run {self.run_number}, cycle {self.cycle}:'

        # setup raise or warn
        if quiet:
            def warn(error, message):
                return False
        elif raise_error:
            def warn(error, message):
                raise error(message)
        else:
            def warn(error, message):
                tqdm.write(message)
                return False

        ## overall data quality checks ---------------------------------------

        # beam data exists
        if len(self.beam1a_current_uA) == 0:
            return warn(BeamError, f'{msg} No 1A beam data saved')

        if len(self.beam1u_current_uA) == 0:
            return warn(BeamError, f'{msg} No 1U beam data saved')

        # total duration
        if self.cycle_stop - self.cycle_start <= 0:
            return warn(DataError, f'{msg} Cycle duration nonsensical: {self.cycle_stop - self.cycle_start} s')

        # valve states
        if not self.cycle_param.valve_states.any().any():
            return warn(ValveError, f'{msg} No valves operated')

        # check if period duration exceeds cycle duration
        expected_duration = self.cycle_param.period_durations_s.sum()
        actual_duration = self.cycle_stop - self.cycle_start
        if expected_duration > actual_duration:
            return warn(DataError, f'{msg} cycle duration ({actual_duration:.1f} s) shorter than sum of periods ({expected_duration:.1f} s)')

        # drop cycles where the 1A beam drops to zero during any time in the cycle
        if any(self.beam1a_current_uA < self.DATA_CHECK_THRESH['beam_min_current']):
            return warn(BeamError, f'{msg} 1A current dropped below {self.DATA_CHECK_THRESH["beam_min_current"]} uA')

        # drop cycles where the 1A beam drops to zero within 5 s of the cycle starting
        if self.cycle > 0:
            cyc_last = self._run[self.cycle-1]
            current = cyc_last.beam1a_current_uA
            idx = current.index > self.cycle_start-20
            if any(current.loc[idx] < self.DATA_CHECK_THRESH["beam_min_current"]):
                return warn(BeamError, f'{msg} 1A current dropped below {self.DATA_CHECK_THRESH["beam_min_current"]} uA within 20 seconds of the cycle starting')

        return True

    def get_nhits(self, detector):
        """Get number of ucn hits

        Args:
            detector (str): Li6|He3
        """
        return self._run._get_nhits(detector, cycle=self.cycle)

    def get_period(self, period=None):
        """Return a copy of this object, but trees are trimmed to only one period.

        Notes:
            * This process converts all objects to dataframes
            * Must be called for a single cycle only
            * Equivalent to indexing style: `cycle[period]`

        Args:
            period (int): period number, if None, get all periods
            cycle (int|None) if cycle not specified then specify a cycle

        Returns:
            run:
                if period > 0: a copy of this object but with data from only one period.
                if period < 0 | None: a list of copies of this object for all periods for a single cycle

        Example:
            ```python
            >>> cycle = run[0]
            >>> cycle.get_period(0)
            run 1846 (cycle 0, period 0):
                comment            cycle_stop         period_start       shifters           tfile
                cycle              experiment_number  period_stop        start_time         year
                cycle_param        month              run_number         stop_time
                cycle_start        period             run_title          supercycle
            ```
        """

        # get all periods
        if period is None or period < 0:
            nperiods = self.cycle_param.nperiods
            return applylist(map(self.get_period, range(nperiods)))
        elif period in self._perioddict.keys():
            return self._perioddict[period]
        else:
            self._perioddict[period] = ucnperiod(self, period)
            return self._perioddict[period]

    def shift_timing(self, dt):
        """Shift all periods by a constant time, maintaining the period durations.
        This shifts the cycle start time and shortens the cycle, potentially creating gaps between cycles

        Args:
            dt (float): time in seconds to add to each period start/end time

        Example:
            ```python
                # this avoids recomputing the histograms each iteration
                dt = [cyc.get_time_shift('Li6', 2, 50) if cyc[2].period_dur > 0 else 0 for cyc in run]
                for i, t in enumerate(dt):
                    run[i].shift_timing(t)
            ```

        Notes:
            * This function makes use of `ucnrun._modify_ptiming`, which resets all saved histograms and hits
        """
        for period in self:
            period.modify_timing(dt, 0)
        self[-1].modify_timing(0, dt)



# Base class for ucnrun, ucncycle, and ucnperiod
# Derek Fujimoto
# Oct 2024

from rootloader import ttree
from .exceptions import *
from .datetime import to_datetime
from .applylist import applylist
import ucndata.constants as const
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from tqdm import tqdm

class ucnbase(object):
    """UCN run data. Cleans data and performs analysis

    Args:
        run (int|str): if int, generate filename with self.datadir
            elif str then run is the path to the file
        header_only (bool): if true, read only the header

    Attributes:
        comment (str): comment input by users
        cycle (int|none): cycle number, none if no cycle selected
        cycle_param (attrdict): cycle parameters
        experiment_number (str): experiment number input by users
        month (int): month of run start
        run_number (int): run number
        run_title (str): run title input by users
        shifter (str): experimenters on shift at time of run
        start_time (str): start time of the run
        stop_time (str): stop time of the run
        supercycle (int|none): supercycle number, none if no cycle selected
        tfile (tfile): stores tfile raw readback
        year (int): year of run start

    Notes:
        Can access attributes of tfile directly from top-level object
        Need to redefine the values if you want non-default
        behaviour
        Object is indexed as [cycle, period] for easy access to sub time frames
    """

    # path to the directory which contains the root files
    datadir = "/data3/ucn/root_files"

    # timezone for datetime conversion
    timezone = 'America/Vancouver'

    # cycle times finding mode order
    cycle_times_mode = ['matched', 'li6', 'he3', 'sequencer', 'beamon']

    # filter what trees and branches to load in each file. If unspecified then load the whole tree
    # treename: (filter, columns). See [rootloader documentation](https://github.com/ucn-triumf/rootloader/blob/main/docs/rootloader/ttree.md#ttree-1) for details
    tree_filter = {}

    # detector tree names
    DET_NAMES = {'He3':{'hits':         'UCNHits_He3',
                        'hits_orig':    'UCNHits_He3', # as saved in the root file
                        'charge':       'He3_Charge',
                        'rate':         'He3_Rate',
                        'transitions':  'RunTransitions_He3',
                        'hitsseq':      'hitsinsequence_he3',
                        'hitsseqcumul': 'hitsinsequencecumul_he3',
                        },
                 'Li6':{'hits':         'UCNHits_Li6',
                        'hits_orig':    'UCNHits_Li-6', # as saved in the root file
                        'charge':       'Li6_Charge',
                        'rate':         'Li6_Rate',
                        'transitions':  'RunTransitions_Li6',
                        'hitsseq':      'hitsinsequence_li6',
                        'hitsseqcumul': 'hitsinsequencecumul_li6',
                        },
                }

    # make these trees dataframes, ok since they're small - also resets the index
    DATAFRAME_TREES = ('CycleParamTree', 'RunTransitions_He3', 'RunTransitions_Li6')

    # needed slow control trees: for checking data quality
    SLOW_TREES = ('BeamlineEpics', 'SequencerTree', 'LNDDetectorTree')

    # EPICS trees to group into a single slow control tree
    EPICS_TREES = [ 'BeamlineEpics',        'UCN2EpPha5Last',   'UCN2EpicsPhase2B',
                    'UCN2EpPha5Oth',        'UCN2EpicsPhase3',  'UCN2EpPha5Pre',
                    'UCN2EpicsPressures',   'UCN2EpPha5Tmp',    'UCN2EpicsTemperature',
                    'UCN2Epics',            'UCN2Pur',          'UCN2EpicsOthers',
                    'UCN2EpLD2'
                  ]

    # data thresholds for checking data
    DATA_CHECK_THRESH = {'beam_min_current': 0.1, # uA
                         'pileup_cnt_per_ms': 10, # if larger than this, then pileup and delete
                         'pileup_within_first_s': 1, # time frame for pileup in each period
                        }

    def __iter__(self):
        # setup iteration
        self._iter_current = 0
        return self

    def _get_beam_duration(self, on=True):
        # Get beam on/off durations

        # get needed info
        cycle_times = self.cycle_param.cycle_times

        try:
            beam = self.tfile.BeamlineEpics
        except AttributeError:
            raise MissingDataError("No saved ttree named BeamlineEpics")

        # setup storage
        beam_dur = []
        epics_val = 'B1V_KSM_RDBEAMON_VAL1' if on else 'B1V_KSM_RDBEAMOFF_VAL1'

        # get as dataframe
        if isinstance(beam, ttree):
            beam = beam[epics_val].to_dataframe()

        # get durations closest to cycle start time
        for start in cycle_times.start:

            start_times = abs(beam.index.values - start)
            durations = beam.values*const.beam_bucket_duration_s
            idx = np.argmin(start_times)
            beam_dur.append(durations[idx])

        out = pd.Series(beam_dur, index=cycle_times.index)
        out.index.name = cycle_times.index.name

        if len(out) == 1:
            return float(out.values[0])

        # select single cycle
        if hasattr(self, 'cycle'):
            out = out[self.cycle]

        return out

    def apply(self, fn_handle):
        """Apply function to each cycle

        Args:
            fn_handle (function handle): function to be applied to each cycle

        Returns:
            applylist: output of the function

        Example:
            ```python
            >>> run.apply(lambda x: x.cycle)
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
            ```
        """
        return applylist([fn_handle(c) for c in self])

    def get_hits_array(self, detector):
        """Get times of ucn hits as a numpy array

        Args:
            detector (str): one of the keys to `self.DET_NAMES`

        Returns:
            np.array: array of timestamps corresponding to an UCN hit. Note that this returns all events in the case where ucn_only=False

        Example:
            ```python
            >>> run.get_hits_array('Li6')
            array([1.75016403e+09, 1.75016405e+09, 1.75016405e+09, ...,
                   1.75016479e+09, 1.75016479e+09, 1.75016479e+09])
            ```
        """

        # check input
        if detector not in self.DET_NAMES.keys():
            raise KeyError(f'Detector input "{detector}" not one of {tuple(self.DET_NAMES.keys())}')

        # get the tree
        tree = self.tfile[self.DET_NAMES[detector]['hits']]

        return tree.tUnixTimePrecise.to_dataframe().index.values

    def get_hits_histogram(self, detector, bin_ms=100, as_datetime=False):
        """Get histogram of UCNHits ttree times

        Args:
            detector (str): Li6|He3
            bin_ms (int): histogram bin size in milliseconds
            as_datetime (bool): if true, convert bin_centers to datetime objects

        Returns:
            rootloader.th1: histogram object

        Example:
            ```python
            >>> run.get_hits_histogram('He3')
            TH1D: "HisttUnixTimePrecise", 557053 entries, sum = 557053.0

            # quick plotting with timestamps
            >>> run.get_hits_histogram('He3', as_datetime=True).plot()

            # get result as a dataframe
            >>> run.get_hits_histogram('He3').to_dataframe()
                    tUnixTimePrecise  Count  Count error
            0         1.750164e+09    0.0     0.000000
            1         1.750164e+09    1.0     1.000000
            2         1.750164e+09    0.0     0.000000
            3         1.750164e+09    0.0     0.000000
            4         1.750164e+09    0.0     0.000000
            ...                ...    ...          ...
            7528      1.750165e+09    0.0     0.000000
            7529      1.750165e+09    0.0     0.000000
            7530      1.750165e+09  415.0    20.371549
            7531      1.750165e+09  507.0    22.516660
            7532      1.750165e+09  509.0    22.561028

            [7533 rows x 3 columns]
            ```
        """

        # check input
        if detector not in self.DET_NAMES.keys():
            raise KeyError(f'Detector input "{detector}" not one of {tuple(self.DET_NAMES.keys())}')

        # check precomputed data
        if self._run._hits_hist['detector'] == detector and \
           self._run._hits_hist['bin_ms'] == bin_ms:

            hist = self._run._hits_hist['hist']

        else:
            # get data
            tree = self._run.tfile[self.DET_NAMES[detector]['hits']]

            # histogram
            hist = tree.hist1d('tUnixTimePrecise', step=bin_ms/1000)
            self._run._hits_hist['detector'] = detector
            self._run._hits_hist['bin_ms'] = bin_ms
            self._run._hits_hist['hist'] = hist

        # make a copy
        hist = hist.copy()

        # trim axes
        if hasattr(self, 'period_start'):
            start = self.period_start
            stop = self.period_stop
        elif hasattr(self, 'cycle_start'):
            start = self.cycle_start
            stop = self.cycle_stop
        else:
            start = 0
            stop = 0

        if start > 0:
            idx = (hist.x >= start) & (hist.x < stop)
            hist.x = hist.x[idx]
            hist.y = hist.y[idx]
            hist.dy = hist.dy[idx]
            hist.sum = np.sum(hist.y)
            hist.entries = int(hist.sum)
            hist.nbins = len(hist.x)


        # to datetime
        if as_datetime:
            hist.x = to_datetime(hist.x)

        return hist

    def plot_psd(self, detector='Li6', cut=None, cmap='RdBu'):
        """Calculate PSD as (QLong-QShort)/QLong, draw as a grid, 2D histograms

        Args:
            detector (str): Li6|He3, select from which detector the data comes from
            cut (tuple): lower left corner of box cut (QLong, PSD). If not none then draw
            cmap (str): [matplotlib color map](https://matplotlib.org/stable/users/explain/colors/colormaps.html) To reverse the order of the colormap append "_r" to the end of the string
        """

        # get the data from the tree
        tree = self.tfile[self.DET_NAMES[detector]['hits']].reset()
        tree.set_filter("tUnixTimePrecise > 15e8", inplace=True)

        # calculate new psd
        tree.set_filter("tChargeL > 0", inplace=True)
        tree.set_filter("tChargeL < 5e3", inplace=True)

        # Li detector figures
        fig, axes = plt.subplots(nrows=3, ncols=3, sharex=True, sharey=True,
                                layout='constrained',
                                figsize=(10,8))
        axes = np.concat(axes)

        hist_edges = []
        max_chargeL = tree.tChargeL.max()

        # histogram
        for i in tqdm(range(9), desc='Generating Histograms', leave=False):

            # hist single channel
            treech = tree.set_filter(f'tChannel=={i}')
            hist = treech.hist2d('tChargeL', 'tPSD', xstep=5, ystep=0.01)

            # reconstruct histogram edges
            xedge = hist.x
            yedge = hist.y

            xedge = np.append(xedge, xedge[-1] + xedge[1]-xedge[0])
            yedge = np.append(yedge, yedge[-1] + yedge[1]-yedge[0])

            xedge -= (xedge[1]-xedge[0])/2
            yedge -= (yedge[1]-yedge[0])/2

            # save
            hist_edges.append((hist.z, xedge, yedge))

        # get max
        max_bin_count = np.max([np.max(h[0]) for h in hist_edges])


        # draw
        for i in range(9):
            ax = axes[i]
            hist, xbins, ybins = hist_edges[i]
            c = ax.pcolormesh(xbins, ybins, hist,
                              cmap=cmap,
                              vmin=0, vmax=max_bin_count/10)
            if i in (2, 5, 8):
                plt.gcf().colorbar(c, ax=ax)
            ax.set_title(f'CH {i}', fontsize='x-small')

            if cut is not None:
                dx = max_chargeL - cut[0]
                dy = 1 - cut[1]
                rec = Rectangle(cut, dx, dy, color='white', fc='none', lw=1)
                ax.add_patch(rec)

        fig.suptitle(f'Run {self.run_number}')
        fig.supxlabel(r'$Q_{\mathrm{Long}}$')
        fig.supylabel(r'$(Q_{\mathrm{Long}}-Q_{\mathrm{Short}})/Q_{\mathrm{Long}}$')

    def trigger_edge(self, detector, thresh, bin_ms=10, rising=True):
        """Detect period start time based on a rising or falling edge
        Args:
            detector (str): Li6|He3
            thresh (float): calculate cycle start time shift based on edge detection passing through this level
            bin_ms (int): histogram bin size in milliseconds
            rising (bool): if true do rising edge, else do falling edge

        Returns:
            np.array: the times at which the rising or falling edge passes through the threshold

        Example:
            ```python
                dt = [cyc[2].get_start_edge('Li6', 50) if cyc[2].period_dur > 0 else 0 for cyc in run]
            ```
        """
        searched = 1 if rising else -1

        # get histogram
        hist = self.get_hits_histogram(detector, bin_ms=bin_ms)
        t = hist.x
        n = hist.y

        # check that the histo has data
        if len(t) == 0:
            raise RuntimeError('Histogram has no data')

        # edge detection using convolution
        # https://stackoverflow.com/questions/50365310/python-rising-falling-edge-oscilloscope-like-trigger
        sign = n >= thresh
        pos = np.where(np.convolve(sign, [1, -1]) == searched)[0]

        # check that positions are found
        if len(pos) == 0:
            return None

        # trim positions which are too long
        pos[pos == len(sign)] -= 1

        return t[pos]

    # quick access properties
    @property
    def beam1a_current_uA(self):
        """Get beamline 1A current in uA (micro amps)

        Returns:
            pd.Series: indexed by timestamps, current in uA
        """
        if type(self.tfile.BeamlineEpics) is pd.DataFrame:
            return self.tfile.BeamlineEpics.B1_FOIL_ADJCUR
        else:
            return self.tfile.BeamlineEpics.B1_FOIL_ADJCUR.to_dataframe()

    @property
    def beam1u_current_uA(self):
        """Get beam current in uA (micro amps)

        Returns:
            pd.Series: indexed by timestamps, current in uA

        Notes:
            Beam current defined as `B1V_KSM_PREDCUR` * `B1V_KSM_BONPRD`

        Example:
            ```python
            >>> run.beam_current_uA
            timestamp
            1572460997    0.0
            1572461002    0.0
            1572461007    0.0
            1572461012    0.0
            1572461017    0.0
                        ...
            1572466463    0.0
            1572466468    0.0
            1572466473    0.0
            1572466478    0.0
            1572466479    0.0
            Length: 1093, dtype: float64
            ```
        """

        # PREDCUR is the predicted current in beamline 1U.
        # PREDCUR is calculated by using the beamline 1V extraction foil current
        # (the current as it leaves the cyclotron) and multiplid by the fraction
        # of beam that is going to the 1U beamline (as opposed to 1A beamline).
        # So if the extraction foil current is 100uA and we are kicking 1 bucket
        # out of 10 buckets to 1U, then PREDCUR will be 10uA

        # BONPRD is a bool, which indicates if there is beam down 1U

        df = self.tfile.BeamlineEpics[['B1V_KSM_PREDCUR', 'B1V_KSM_BONPRD']]
        df = df.to_dataframe()

        # current in the 1U beamline
        return df.B1V_KSM_PREDCUR * df.B1V_KSM_BONPRD

    @property
    def beam_on_s(self):
        """Get the beam-on duration in seconds for each cycle as given by `B1V_KSM_RDBEAMON_VAL1`

        Returns:
            pd.Series: indexed by cycle number, beam-on duration in s

        Example:

            ```python
            >>> run.beam_on_s
            cycle
            0     59.999284
            1     59.999284
            2     59.999284
            3     59.999284
            4     59.999284
            5     59.999284
            6     59.999284
            7     59.999284
            8     59.999284
            9     59.999284
            10    59.999284
            11    59.999284
            12    59.999284
            13    59.999284
            14    59.999284
            15    59.999284
            16    59.999284
            dtype: float64
            ```
        """

        return self._get_beam_duration(on=True)

    @property
    def beam_off_s(self):
        """Get the beam-off duration in seconds for each cycle as given by `B1V_KSM_RDBEAMOFF_VAL1`

        Returns:
            pd.Series: indexed by cycle number, beam-off duration in s

        Example:

            ```python
            >>> run.beam_off_s
            cycle
            0     229.999918
            1     229.999918
            2     229.999918
            3     229.999918
            4     229.999918
            5     229.999918
            6     229.999918
            7     229.999918
            8     229.999918
            9     229.999918
            10    229.999918
            11    229.999918
            12    229.999918
            13    229.999918
            14    229.999918
            15    229.999918
            16    229.999918
            dtype: float64
            ```
        """
        return self._get_beam_duration(on=False)





# Open and analyze UCN data for a whole run
# Derek Fujimoto
# June 2024

from rootloader import tfile, ttree, attrdict
from .exceptions import *
from .applylist import applylist
from .ucnbase import ucnbase
from .ttreeslow import ttreeslow
from .ucncycle import ucncycle
from .datetime import to_datetime
import itertools, warnings, os, ROOT

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from collections.abc import Iterable
from tqdm import tqdm

ROOT.gROOT.SetBatch(1)

# patch warnings
def new_format(message, category, filename, lineno, line):
    filename = os.path.basename(filename)
    return f'\n{filename}:{lineno}: {category.__name__}: {message}\n'

warnings.formatwarning = new_format

class ucnrun(ucnbase):
    """UCN run data. Cleans data and performs analysis

    Args:
        run (int|str): if int, generate filename with self.datadir
            elif str then run is the path to the file
        ucn_only (bool): if true set filter tIsUCN==1 on all hit trees

    Attributes:
        comment (str): comment input by users
        cycle (int|none): cycle number, none if no cycle selected
        cycle_param (attrdict): cycle parameters from sequencer settings
        experiment_number (str): experiment number input by users
        month (int): month of run start
        run_number (int): run number
        run_title (str): run title input by users
        shifter (str): experimenters on shift at time of run
        start_time (str): start time of the run
        stop_time (str): stop time of the run
        supercycle (int|none): supercycle number, none if no cycle selected
        tfile (tfile): stores tfile raw readback
        year (int): year of run start

    Notes:
        * Can access attributes of tfile directly from top-level object
        * Need to define the values if you want non-default behaviour
        * Object is indexed as [cycle, period] for easy access to sub time frames

        Cycle param contents

        `nperiods`: Number of periods in each cycle
        `nsupercyc`: Number of supercycles contained in the run
        `enable`: Enable status of the sequencer
        `inf_cyc_enable`: Enable status of infinite cycles
        `cycle`: Cycle ID numbers
        `supercycle`: Supercycle ID numbers
        `valve_states`: Valve states in each period and cycle
        `period_end_times`: End time of each period in each cycle in epoch time
        `period_durations_s`: Duration in sections of each period in each cycle
        `ncycles`: Number of total cycles contained in the run
        `filter`: A list indicating how we should filter cycles. More on that in [filters](filters.md)
        `cycle_time`: The start and end times of each cycle

    Examples:

        Loading runs
        ```python
        from ucndata import ucnrun, settings
        # load from filename
        ucnrun('/path/to/file/ucn_run_00002684.root')
        # load from run number
        ucnrun.datadir = '/path/to/file/'
        ucnrun(2684)
        ```

        Slicing
        ```python
        from ucndata import ucnrun
        run = ucnrun(2684)
        run[0, 0]   # cycle 0, period 0
        run[:]      # all cycles, no distinction on period
        run[:, 0]   # get all cycles, period 0
        run[0, :]   # cycle 0, all periods
        run[4:7, 2] # cycles 4, 5, 6, period 2
        ```

        Get beam properties
        ```python
        from ucndata import ucnrun
        run = ucnrun(2684)
        run.beam_current_uA # beam current in uA
        run.beam_on_s       # beam duration on in s
        run.beam_off_s      # beam duration off in s
        ```

        Draw hits
        ```python
        import matplotlib.pyplot as plt
        from ucndata import ucnrun
        run = ucnrun(2684)

        # draw all hits in the file
        run.get_hits_histogram('Li6').plot()

        # draw hits in each cycle
        for cycle in run:
            cycle.get_hits_histogram('Li6').plot(label=cycle.cycle)

        # adjust the timing of cycles and periods
        run._modify_ptiming(cycle=0, period=0, dt_start_s=1, dt_start_s=0)

        # inspect the data: draw a figure with hits histogram, beam current, and optional slow control data
        run.inspect('Li6', bin_ms=100, xmode='dur')
        ```
    """

    def __init__(self, run, ucn_only=True):

        # check if copying
        if run is None:
            return

        # make filename from defaults
        elif isinstance(run, (int, float, np.float64, np.int64)):
            filename = os.path.join(self.datadir, f'ucn_run_{int(run):0>8d}.root')

        # fetch from specified path
        elif isinstance(run, str):
            filename = run

        else:
            raise TypeError(f'Unknown input type "{type(run)}" for run')

        self.path = os.path.abspath(filename)

        # read
        self.tfile = tfile(filename, empty_ok=False, quiet=True,
                            key_filter=self.keyfilter)
        head = self.tfile['header'].to_dataframe()

        # reformat header and move to top level
        for k, val in head.items():
            setattr(self, k.replace(' ', '_').lower(), val.loc[0])

        if type(self.run_number) is pd.Series:
            self.run_number = int(float(self.run_number[0]))
        else:
            self.run_number = int(float(self.run_number))

        # set other header items
        date = pd.to_datetime(self.start_time)
        self.year = date.year
        self.month = date.month

        # reformat tfile branch names to remove spaces
        for key in tuple(self.tfile.keys()):
            if ' ' in key or '-' in key:
                self.tfile[key.replace(' ', '_').replace('-','')] = self.tfile[key]
                del self.tfile[key]

        # some trees should be dataframes by default
        for name in self.DATAFRAME_TREES:
            if name in self.tfile.keys():
                self.tfile[name] = self.tfile[name].to_dataframe()
                self.tfile[name].reset_index(inplace=True, drop=True)

        # set cycle parameters
        self._get_cycle_param()

        # get cycle times
        if hasattr(self, 'cycle_param'):
            cycle_failed = False
            if isinstance(self.cycle_times_mode, str):
                self.cycle_times_mode = [self.cycle_times_mode]

            for mode in self.cycle_times_mode:
                try:
                    self.set_cycle_times(mode=mode)
                except (AttributeError, IndexError):
                    warnings.warn(f'Run {self.run_number}: Unable to set cycle times. SequencerTree must exist and have entries.',
                                MissingDataWarning)
                    break
                except (KeyError, CycleError):
                    print(f'Run {self.run_number}: Cycle time detection mode {mode} failed')
                    cycle_failed = True
                else:
                    break

            if cycle_failed:
                print(f'Run {self.run_number}: Set cycle times based on {mode} detection mode')

        # setup tree filters
        for detector in self.DET_NAMES.keys():
            try:
                tree = self.tfile[self.DET_NAMES[detector]['hits']]
            except KeyError:
                continue

            # purge bad timestamps
            tree.set_filter('tUnixTimePrecise > 15e8', inplace=True)

            # filter on ucn hits
            if ucn_only:
                tree.set_filter('tIsUCN>0', inplace=True)


        # make slow control tree
        self.epics = ttreeslow((self.tfile[name] for name in self.EPICS_TREES),
                               parent=self)

        # store fetched cycles
        self._cycledict = dict()

        # store fetched histogram
        self._hits_hist = {'detector':None,
                           'bin_ms': None,
                           'hist':None}

        # histograms with edges set by period and cycle timings
        self._nhits = None

        # pointer to self for cycles and periods
        self._run = self

    def __next__(self):
        # permit iteration over object like it was a list

        if self._iter_current < self.cycle_param.ncycles:

            # skip elements that should be filtered
            if self.cycle_param.filter is not None:

                # skip
                while not self.cycle_param.filter[self._iter_current]:
                    self._iter_current += 1

                    # end condition
                    if self._iter_current >= self.cycle_param.ncycles:
                        raise StopIteration()

            # iterate
            cyc = self[self._iter_current]
            self._iter_current += 1
            return cyc

        # end of iteration
        else:
            raise StopIteration()

    def __repr__(self):
        klist = [d for d in self.__dict__.keys() if d[0] != '_']
        if klist:

            # sort without caps
            klist.sort(key=lambda x: x.lower())

            # get number of columns based on terminal size
            maxsize = max((len(k) for k in klist)) + 2
            terminal_width = os.get_terminal_size().columns
            ncolumns = int(np.floor(terminal_width / maxsize))
            ncolumns = min(ncolumns, len(klist))

            # split into chunks
            needed_len = int(np.ceil(len(klist) / ncolumns)*ncolumns) - len(klist)
            klist = np.concatenate((klist, np.full(needed_len, '')))
            klist = np.array_split(klist, ncolumns)

            # print
            s = f'run {self.run_number}:\n'
            for key in zip(*klist):
                s += '  '
                s += ''.join(['{0: <{1}}'.format(k, maxsize) for k in key])
                s += '\n'
            return s
        else:
            return self.__class__.__name__ + "()"

    def __getitem__(self, key):
        # get cycle or period based on slicing indexes

        # get a single key
        if isinstance(key, (np.integer, int)):

            if key < 0:
                key = self.cycle_param.ncycles + key

            if key > self.cycle_param.ncycles:
                raise IndexError(f'Run {self.run_number}: Index larger than number of cycles ({self.cycle_param.ncycles})')

            return self.get_cycle(key)

        # slice on cycles
        if isinstance(key, slice):
            cycles = self.get_cycle()[:self.cycle_param.ncycles]

            # no filter
            if self.cycle_param.filter is None or all(self.cycle_param.filter):
                cyc = cycles[key]

            # yes filter
            else:

                # fetch the filter and slice in the same way as the return value
                cfilter = self.cycle_param.filter[key]

                # fetch cycles and slice, then apply filter
                cyc = np.array(cycles[key])
                cyc = cyc[cfilter]

            return applylist(cyc)

        # slice on periods
        if isinstance(key, tuple):
            cycles = self[key[0]]
            if isinstance(cycles, (np.ndarray, applylist, list)):
                return applylist([c[key[1]] for c in cycles])
            else:
                return cycles[key[1]]

        raise IndexError(f'Run {self.run_number} given an unknown index type ({type(key)})')

    def __len__(self):
        return self.cycle_param.ncycles

    def _get_cycle_param(self):
        # set self.cycle_param dict

        # reformat cycle param tree
        paramtree = self.tfile.CycleParamTree
        self.cycle_param = {'nperiods': paramtree.nPeriods[0],
                            'nsupercyc': paramtree.nSuperCyc[0],
                            'enable': bool(paramtree.enable[0]),
                            'inf_cyc_enable': bool(paramtree.infCyclesEnable[0]),
                            'ncycles': 1
                            }
        self.cycle_param = attrdict(self.cycle_param)

        # cycle filter
        self.cycle_param['filter'] = None

        # setup cycle paramtree array outputs from transition trees
        treelist = []
        for detector in self.DET_NAMES.values():
            if detector['transitions'] in self.tfile.keys():
                treelist.append(self.tfile[detector['transitions']])

        # check if trees exist
        if not treelist:
            warnings.warn(f'Run {self.run_number}: no detector transition tree, cannot fully set up cycle_param', MissingDataWarning)

            # number of cycles, periods
            self.cycle_param['ncycles'] = 1
            self.cycle_param['nperiods'] = 1

            # cycle and supercycle indices
            self.cycle_param['cycle'] = pd.Series([0], index=[0])
            self.cycle_param['supercycle'] = pd.Series([0], index=[0])

            self.cycle_param['cycle'].name = 'cycleIndex'
            self.cycle_param['supercycle'].name = 'superCycleIndex'

            return

        # get longest
        treelen = [len(t) for t in treelist]
        tree = treelist[np.argmax(treelen)]

        # cycle and supercycle indices
        self.cycle_param['cycle'] = tree['cycleIndex'].astype(int)
        self.cycle_param['supercycle'] = tree['superCycleIndex'].astype(int)

        # convert the array in each cell into a dataframe
        def item_to_df(x):
            s = pd.DataFrame(np.array(x).copy(), index=np.arange(len(x))+1)
            return s

        # valve states -------------------------------------------------------
        df = tree[[col for col in tree.columns if 'valveState' in col]]
        col_map = {col:int(col.replace("valveStatePeriod", "")) for col in df.columns}
        df = df.rename(columns=col_map)
        df.columns.name = 'period'
        df.reset_index(inplace=True, drop=True)
        df.index.name = 'cycle_idx'

        # valve states should not change across cycles
        df2 = df.loc[0]
        df2 = pd.concat([item_to_df(df2[period]) for period in df2.index], axis='columns')

        # rename columns and index
        df2.columns = np.arange(len(df2.columns))
        df2.columns.name = 'period'
        df2.index.name = 'valve'
        self.cycle_param['valve_states'] = df2.transpose()

        # period end times ---------------------------------------------------
        df = tree[[col for col in tree.columns if 'cyclePeriod' in col]]
        col_map = {col:int(col.replace("cyclePeriod", "").replace("EndTime", "")) for col in df.columns}

        # rename columns and index
        df = df.rename(columns=col_map)
        df.columns.name = 'period'
        df.index.name = 'cycle'
        self.cycle_param['period_end_times'] = df.transpose()

        # period durations ---------------------------------------------------
        cycle_start = tree.cycleStartTime
        df_diff = df.diff(axis='columns')
        df_diff[0] = df[0] - cycle_start

        self.cycle_param['period_durations_s'] = df_diff.transpose()

        # number of cycles
        self.cycle_param['ncycles'] = len(df.index)

    def _get_nhits(self, detector, cycle=None, period=None):
        # Get number ucn hits

        # Args:
        #     detector (str): Li6|He3
        #     cycle (int): cycle number, if None compute for whole run
        #     period (int): period number, if None compute for whole cycle, if cycle is not also None

        # Notes:
        #     Because of how RDataFrame works it is better to compute a histogram whose bins correspond to the period or cycle start/end times than to set a filter and get the resulting tree size.
        #     The histogram bin quantities is saved as self._nhits or self._nhits_cycle
        #     Both ucncycle and ucnperiod classes call this function to get the counts

        # get hits tree
        tree = self.tfile[self.DET_NAMES[detector]['hits']]

        # get hits for run
        if cycle is None and period is None:
            return tree.size

        else:

            # make hits histogram
            if self._nhits is None:

                # use period and cycle start and end times as bin edges
                edges = []
                for cyclei in range(self.cycle_param.ncycles):

                    # get cycle start and end
                    cycle_start = self.cycle_param.cycle_times.start[cyclei]
                    period_ends = self.cycle_param.period_end_times[cyclei]

                    # shorten periods that extend past the end of the cycle (edge case)
                    if cyclei < self.cycle_param.ncycles-1:
                        next_cycle_start = self.cycle_param.cycle_times.start[cyclei+1]
                    else:
                        next_cycle_start = self.cycle_param.cycle_times.stop.iloc[-1]

                    period_ends[period_ends > next_cycle_start] = next_cycle_start

                    # save to list
                    edges.append(cycle_start)
                    edges.extend(list(period_ends))

                edges = np.append(edges, self.cycle_param.cycle_times.stop.iloc[-1])
                edges = np.append(edges, self.cycle_param.cycle_times.stop.iloc[-1]) # need duplicate end bin for some reason

                self._nhits = tree.hist1d('tUnixTimePrecise', edges=edges).y[1:]

            # get hits for cycle
            if period is None:
                nperiodsp1 = len(self.cycle_param.period_end_times)+1
                return int(self._nhits[cycle*nperiodsp1:(cycle+1)*nperiodsp1].sum())

            # get hits for period
            else:
                return int(self._nhits[cycle*(len(self.cycle_param.period_end_times)+1)+period])

    def _modify_ptiming(self, cycle, period, dt_start_s=0, dt_stop_s=0, update_duration=True):
        # Change start and end times of periods

        # Args:
        #     cycle (int): cycle id number
        #     period (int): period id number
        #     dt_start_s (float): change to the period start time
        #     dt_stop_s (float): change to the period stop time
        #     update_duration (bool): if true, update period durations dataframe

        # Notes:
        #     * as a result of this, cycles may overlap or have gaps
        #     * periods are forced to not overlap and have no gaps
        #     * cannot change cycle end time, but can change cycle start time
        #     * this function resets all saved histgrams and hits

        # get cycle parameters
        cycpar = self.cycle_param

        # start times - prevent unnecessary recursion
        if dt_start_s != 0:

            # adjust cycle start
            if period == 0:
                cycpar.cycle_times.loc[cycle, 'start'] += dt_start_s

            # period start time adjustment
            else:
                cycpar.period_end_times.loc[period-1, cycle] += dt_start_s

            # recursively adjust adjacent size zero periods
            if period-1 > 0 and cycpar.period_durations_s.loc[period-1, cycle] == 0:
                self._modify_ptiming(cycle, period-1,
                                    dt_start_s = dt_start_s,
                                    dt_stop_s  = 0,
                                    update_duration = False)

        # stop times - prevent unnecessary recursion
        if dt_stop_s != 0:

            # period end time adjustment
            cycpar.period_end_times.loc[period, cycle] += dt_stop_s

            # force periods to stay within cycle bounds
            cycend = cycpar.cycle_times.loc[cycle, 'stop']
            perend = cycpar.period_end_times.loc[period, cycle]
            cycpar.period_end_times.loc[period, cycle] = min(perend, cycend)

            # recursively adjust adjacent size zero periods
            try:
                if cycpar.period_durations_s.loc[period+1, cycle] == 0:
                    self._modify_ptiming(cycle, period+1,
                                        dt_start_s = 0,
                                        dt_stop_s  = dt_stop_s,
                                        update_duration = False)

            # period not in durations dataframe
            except KeyError:
                pass

        # adjust period and cycle durations
        if update_duration:

            start = cycpar.cycle_times['start']
            stop = cycpar.cycle_times['stop']

            df = cycpar.period_end_times
            df_diff = df.diff()
            df_diff.loc[0] = df.loc[0] - start
            cycpar.period_durations_s = df_diff

            cycpar.cycle_times.loc[cycle, 'duration (s)'] = (stop-start).loc[0]

        # set in memory
        self.cycle_param = cycpar

        # remove saved cycles to account for updated limits
        if cycle in self._cycledict.keys():
            del self._cycledict[cycle]

        # reset histogram for number of hits
        self._nhits = None

        # reset all saved histograms
        for key in self._hits_hist.keys():
            self._hits_hist[key] = None

    def check_data(self, raise_error=False):
        """Run some checks to determine if the data is ok.

        Args:
            raise_error (bool): if true, raise an error if check fails, else return false

        Returns:
            bool: true if check passes, else false.

        Notes:
            * Do the self.SLOW_TREES exist and have entries?
            * Are there nonzero counts in UCNHits?

        Example:
            ```python
            >>> run.check_data()
            True
            ```
        """

        # check some necessary data trees
        for tree in self.SLOW_TREES:

            msg = None

            # does tree exist?
            if tree not in self.tfile.keys():
                msg = f'Missing ttree "{tree}" in run {self.run_number}'

            # does tree have entries?
            elif len(self.tfile[tree]) == 0:
                msg = f'Zero entries found in "{tree}" ttree in run {self.run_number}'

            # raise error or return
            if msg is not None:
                if raise_error:
                    raise MissingDataError(msg)
                else:
                    print(msg)
                    return False

        for name, det in self.DET_NAMES.items():

            # check for nonzero counts
            if not self.tfile[det['hits']].tIsUCN.any():
                msg = f'No UCN hits in "{name}" ttree in run {self.run_number}'

            # check if sequencer was enabled but no run transitions
            elif any(self.tfile.SequencerTree.sequencerEnabled):

                if len(self.tfile[det['transitions']]) == 0:
                    msg = 'No cycles found in run {self.run_number}, although sequencer was active'

            # raise error or return
            if msg is not None:
                if raise_error:
                    raise MissingDataError(msg)
                else:
                    tqdm.write(msg)
                    return False

        return True

    def draw_cycle_times(self, ax=None, xmode='datetime', do_legend=False):
        """Draw cycle start times as thick black lines, period end times as dashed lines

        Args:
            ax (plt.Axes): axis to draw in, if None, draw in current axes
            xmode (str): datetime|duration|epoch
            do_legend (bool): if true draw legend colors for period numbers

        Notes:
            Assumed periods:    0 - irradiation
                                1 - storage
                                2 - count
        """

        # check input
        if all((xmode not in i for i in ('datetime', 'duration', 'epoch'))):
            raise RuntimeError('xmode must be one of datetime|duration|epoch')

        # get axis to draw in
        if ax is None:
            ax = plt.gca()

        # run start time
        if xmode in 'duration':
            run_start = self.cycle_param.cycle_times.loc[0, 'start']
        else:
            run_start = 0

        # draw lines
        non_zero_periods = []
        for cyc in self.get_cycle():

            # get x value
            if xmode in 'datetime':
                start = pd.to_datetime(cyc.cycle_start, unit='s')
            else:
                start = cyc.cycle_start - run_start

            # draw
            ax.axvline(start, color='k', ls='-', lw=2)

            for i, per in enumerate(cyc):

                # skip zero length periods
                if per.period_start == per.period_stop:
                    continue

                # get x value
                if xmode in 'datetime':
                    x = pd.to_datetime(per.period_stop, unit='s')
                else:
                    x = per.period_stop - run_start

                # draw
                ax.axvline(x, color=f'C{i}', ls=':', lw=1)
                non_zero_periods.append(i)

            # get cycle text - strikeout if not good
            text = f'Cycle {cyc.cycle}'

            # check if filtered
            if cyc.cycle_param.filter is None or cyc.cycle_param.filter:
                color = 'k'
            else:
                text = '\u0336'.join(text) + '\u0336'
                color = 'r'

            text += ' $\\downarrow$'

            ypos = ax.get_ylim()[1]
            ax.text(start, ypos, text,
                    va='top',
                    ha='left',
                    fontsize='xx-small',
                    color=color,
                    rotation='vertical',
                    clip_on=True,)

        # add periods to legend
        if do_legend:
            handles = [mpatches.Patch(color=f'C{period}', label=f'Period {period}') \
                    for period in np.unique(non_zero_periods)]
            ax.legend(handles=handles)

    def gen_cycle_filter(self, quiet=False):
        """Generate filter array for cycles. Use with self.set_cycle_filter to filter cycles.

        Args:
            quiet (bool): if true don't print or raise exception

        Returns:
            np.array(bool): true if keep cycle, false if discard

        Notes:
            calls `ucncycle.check_data` on each cycle

        Example:
            ```python
            >>> run = ucnrun(2575)
            >>> run.gen_cycle_filter()
            Run 2575, cycle 0: 1A current dropped below 0.1 uA
            Run 2575, cycle 1: 1A current dropped below 0.1 uA

            array([False, False,  True,  True,  True,  True,  True,  True,  True,
                    True,  True,  True])
            ```
        """

        cycles = self.get_cycle()
        iterator = tqdm(cycles, desc=f'Run {self.run_number}: Scanning cycles',
                                leave=False,
                                total=self.cycle_param.ncycles)
        cfilter = [c.check_data(quiet=quiet, raise_error=False) for c in iterator]
        return np.array(cfilter)

    def get_cycle(self, cycle=None):
        """Return a copy of this object, but trees are trimmed to only one cycle.

        Note that this process converts all objects to dataframes

        Args:
            cycle (int): cycle number, if None, get all cycles

        Returns:
            ucncycle:
                if cycle > 0:  ucncycle object
                if cycle < 0 | None: a list ucncycle objects for all cycles

        Examples:
            ```python
            # get single cycle
            >>> run.get_cycle(0)
            run 1846 (cycle 0):
                comment            cycle_start        month              shifters           supercycle
                cycle              cycle_stop         run_number         start_time         tfile
                cycle_param        experiment_number  run_title          stop_time          year

            # get all cycles
            >>> len(run.get_cycle())
            17
            ```
        """

        if cycle is None or cycle < 0:
            ncycles = len(self.cycle_param.cycle_times.index)
            return applylist(map(self.get_cycle, range(ncycles)))
        elif cycle in self._cycledict.keys():
            return self._cycledict[cycle]
        else:
            self._cycledict[cycle] = ucncycle(self, cycle)
            return self._cycledict[cycle]

    def inspect(self, detector='Li6', bin_ms=100, xmode='duration', slow=None):
        """Draw counts and BL1A current with indicated periods to determine data quality

        Args:
            detector (str): detector from which to get the counts from. Li6|He3
            bin_ms (int): histogram bin size in ms
            xmode (str): datetime|duration|epoch
            slow (list|str): name of slow control tree to add in a separate axis, can be a list of names
        """

        # check input
        if all((xmode not in i for i in ('datetime', 'duration', 'epoch'))):
            raise RuntimeError('xmode must be one of datetime|duration|epoch')

        # number of rows in figure
        nrows = 2

        # check input
        if slow is not None:
            if isinstance(slow, str):
                if slow not in self.epics.columns:
                    raise KeyError(f'Input slow ("{slow}") not found in tree "epics"')
                slow = [slow]
            elif isinstance(slow, Iterable):
                for s in slow:
                    if s not in self.epics.columns:
                        raise KeyError(f'Input slow ("{slow}") not found in tree "epics"')
            else:
                raise RuntimeError('Input "slow" must be a string or iterable')

            # extra row for slow control
            nrows += len(slow)

        # make figure
        _, axes = plt.subplots(nrows=nrows, ncols=1, sharex=True,
                        layout='constrained', figsize=(8,10))

        # get current and histogram
        current = self.beam1a_current_uA
        current.sort_index(inplace=True)
        hist = self.get_hits_histogram(detector, bin_ms=bin_ms).to_dataframe()
        hist.set_index('tUnixTimePrecise', inplace=True)
        hist = hist['Count']

        # get slow control
        if slow is not None:
            slow = {s:self.epics[s].to_dataframe() for s in slow}

        # run start time
        run_start = self.cycle_param.cycle_times.loc[0, 'start']

        for cyc in self.get_cycle():
            for i, per in enumerate(cyc):

                # draw current
                cur = current.loc[per.period_start:per.period_stop]

                if len(cur) > 0:
                    if xmode in 'datetime':
                        cur.index = pd.to_datetime(cur.index, unit='s')
                    elif xmode in 'duration':
                        cur.index -= run_start

                    cur.plot(ax=axes[0], color=f'C{i}')

                # draw histogram
                hi = hist.loc[per.period_start:per.period_stop]

                if len(hi) > 0:
                    if xmode in 'datetime':
                        hi.index = pd.to_datetime(hi.index, unit='s')
                    elif xmode in 'duration':
                        hi.index -= run_start

                    hi.plot(ax=axes[1], color=f'C{i}')

                # draw slow control
                if slow is not None:
                    for j, (key, val) in enumerate(slow.items()):
                        v = val.loc[per.period_start:per.period_stop]
                        if len(v) > 0:
                            if xmode in 'datetime':
                                v.index = pd.to_datetime(v.index, unit='s')
                            elif xmode in 'duration':
                                v.index -= run_start
                        v.plot(ax=axes[j+2], color=f'C{i}')

            # draw the rest of the run - current
            cur = current.loc[per.period_stop:cyc.cycle_stop]

            if len(cur) > 0:
                if xmode in 'datetime':
                    cur.index = pd.to_datetime(cur.index, unit='s')
                elif xmode in 'duration':
                    cur.index -= run_start

                cur.plot(ax=axes[0], color=f'k')

            # draw the rest of the run - histogram
            hi = hist.loc[per.period_stop:cyc.cycle_stop]

            if len(hi) > 0:
                if xmode in 'datetime':
                    hi.index = pd.to_datetime(hi.index, unit='s')
                elif xmode in 'duration':
                    hi.index -= run_start

                hi.plot(ax=axes[1], color=f'k')

            # draw slow control
            if slow is not None:
                for i, (key, val) in enumerate(slow.items()):
                    v = val.loc[per.period_stop:cyc.cycle_stop]
                    if len(v) > 0:
                        if xmode in 'datetime':
                            v.index = pd.to_datetime(v.index, unit='s')
                        elif xmode in 'duration':
                            v.index -= run_start
                    v.plot(ax=axes[i+2], color=f'k')

        # draw vertical markers
        for i, ax in enumerate(axes):
            self.draw_cycle_times(ax=ax, xmode=xmode, do_legend=(i==0))

        # plot elements
        axes[0].set_title(self.run_number,fontsize='x-small')
        axes[0].set_ylabel('BL1A Current (uA)')
        axes[1].set_ylabel(f'UCN Counts/{bin_ms/1000:g}s')

        if slow is not None:
            slow_keys = tuple(slow.keys())
            for i, ax in enumerate(axes[2:]):
                ax.set_ylabel(slow_keys[i].split('_')[-2])

        if xmode in 'datetime':
            axes[-1].set_xlabel('')
        elif xmode in 'duration':
            axes[-1].set_xlabel('Time Since Run Start (s)')
        else:
            axes[-1].set_xlabel('Epoch Time')
        axes[1].set_yscale('log')

    def keyfilter(self, name):
        """Don't load all the data in each file, only that which is needed"""

        name = name.replace(' ', '_').lower()

        # reject some keys based on partial matches
        reject_keys = ('v1725', 'v1720', 'v792', 'tv1725', 'charge', 'edge_diff',
                       'pulse_widths', 'iv2', 'iv3', 'rate')
        for key in reject_keys:
            if key in name:
                return False

        return True

    def set_cycle_filter(self, cfilter=None):
        """Set filter for which cycles to fetch when slicing or iterating

        Args:
            cfilter (None|iterable): list of bool, True if keep cycle, False if reject.
                if None then same as if all True

        Returns:
            None: sets self.cycle_param.filter

        Notes:
            Filter is ONLY applied when fetching cycles as a slice or as an iterator. ucnrun.get_cycle() always returns unfiltered cycles.

            Examples where the filter is applied:
                * run[:]
                * run[3:10]
                * run[:3]
                * for c in run: print(c)

            Examples where the filter is not applied:
                * run[2]
                * run.get_cycle()
                * run.get_cycle(2)

        Example:

            ```python
            # check how many cycles are fetched without filter
            >>> len(run[:])
            17

            # apply a filter
            >>> filter = np.full(17, True)
            >>> filter[2] = False
            >>> run.set_cycle_filter(filter)

            # check that cycle 2 is filtered out
            >>> len(run[:])
            16
            >>> for c in run:
                    print(c.cycle)
            0
            1
            3
            4
            5
            6
            7
            8
            9
            10
            11
            12
            13
            14
            15
            16
            ```
        """

        # check input
        cfilter = np.array(cfilter).astype(bool)

        if len(cfilter) != self.cycle_param.ncycles:
            raise RuntimeError(f'Run {self.run_number}: Length of cycle filter ({len(cfilter)}) does not match expected number of cycles ({self.cycle_param.ncycles})')

        # set
        self.cycle_param.filter = cfilter

        # set for already fetched cycles
        for key, cyc in self._cycledict.items():
            cyc.cycle_param.filter = self.cycle_param.filter[key]

    def set_cycle_times(self, mode):
        """Get start and end times of each cycle from the sequencer and save
        into self.cycle_param.cycle_times

        Run this if you want to change how cycle start times are calculated

        Args:
            mode (str): default|matched|sequencer|he3|li6
                if matched: look for identical timestamps in RunTransitions from detectors
                if sequencer: look for inCycle timestamps in SequencerTree
                if he3: use He3 detector cycle start times
                if li6: use Li6 detector cycle start times

        Returns:
            pd.DataFrame: with columns "start", "stop", "offset" and "duration (s)". Values are in epoch time. Indexed by cycle id. Offset is the difference in detector start times: he3_start-li6_start


        Notes:
            - If run ends before sequencer stop is called, a stop is set to final timestamp.
            - If the sequencer is disabled mid-run, a stop is set when disable ocurrs.
            - If sequencer is not enabled, then make the entire run one cycle
            - For matched mode,
                - set run stops as start of next transition
                - set offset as start_He3 - start_Li6
                - set start/stop/duration based on start_He3
            - If the object reflects a single cycle, return from cycle_start, cycle_stop

        Example:
            ```python
            # this calculates new cycle start and end times based on the selected method
            >>> run.set_cycle_times('li6')
            ```
        """

        # check if single cycle
        if hasattr(self, 'cycle'):
            return pd.DataFrame({'start':[self.cycle_start],
                                 'stop':[self.cycle_stop],
                                 'duration (s)': [self.cycle_stop-self.cycle_start],
                                 'offset (s)': [0.0],
                                 'supercycle': [self.supercycle]},
                                 index=[self.cycle])

        # get mode
        mode = mode.lower()

        # get run end time from control trees - used in matched and detector cycles times
        run_stop = -np.inf
        for treename in self.SLOW_TREES:
            try:
                max_time = self.tfile[treename].index.max()

            # tree not found, skip
            except KeyError:
                continue

            run_stop = max((max_time, run_stop))

        # bad end time
        if np.isinf(run_stop):
            raise MissingDataError(f"Run {self.run_number}: Missing slow control trees, cannot find run end time.")

        # get squencer data
        try:
            df = self.tfile.SequencerTree
        except AttributeError:
            df = None
        else:
            if type(df) == ttree:
                df = df.to_dataframe()

        ## if no sequencer, make the whole run a single cycle
        if df is None or not any(df.sequencerEnabled):

            times = {'start': np.inf,
                     'stop':  -np.inf,
                     'supercycle': [0]}

            # use timestamps from slow control trees to determine timestamps
            for treename in self.SLOW_TREES:

                # check for tree
                if treename not in self.tfile.keys():
                    continue

                # get times
                if isinstance(self.tfile[treename], pd.DataFrame):
                    idx = self.tfile[treename].index
                else:
                    idx = self.tfile[treename].to_dataframe().index

                # find min and max
                times['start'] = [min((idx.min(), times['start']))]
                times['stop']  = [max((idx.max(), times['stop']))]
                break

        ## get matched timesteps from He3 and Li6 RunTransitions
        elif mode in 'matched':
            hestart = self.tfile[self.DET_NAMES['He3']['transitions']].cycleStartTime
            listart = self.tfile[self.DET_NAMES['Li6']['transitions']].cycleStartTime
            scycle = self.tfile[self.DET_NAMES['He3']['transitions']].superCycleIndex

            # drop duplicate timestamps
            hestart = hestart.drop_duplicates()
            listart = listart.drop_duplicates()

            # get all possible pairs and sort by time difference
            pairs = sorted(itertools.product(hestart, listart),
                            key = lambda t: abs(t[0] - t[1]))

            # save output
            matchedhe3 = []
            matchedli6 = []

            # go through all possible pairs
            for pair in pairs:

                # if none of the two start times are already in the matched list, add the pair to the matched list
                if pair[0] not in matchedhe3 and pair[1] not in matchedli6:
                    offset = pair[0] - pair[1]

                    # discard if time difference is too large
                    if abs(offset) > 20:
                        raise CycleError(f'He3 cycle start time ({pair[0]}) too distant from Li6 start ({pair[1]}) in run {self.run_number}')
                    else:
                        matchedhe3.append(pair[0])
                        matchedli6.append(pair[1])

            matchedhe3 = np.sort(matchedhe3)
            matchedli6 = np.sort(matchedli6)

            # warnings for unmatched cycles
            unmatched = [t not in matchedhe3 for t in hestart.values]
            if any(unmatched):
                raise CycleError(f'Found no match to He3 cycles at {hestart.values[unmatched]} in run {self.run_number}')

            unmatched = [t not in matchedhe3 for t in listart.values]
            if any(unmatched):
                raise CycleError(f'Found no match to Li6 cycles at {listart.values[unmatched]} in run {self.run_number}')

            # setup output
            times = {'start': matchedhe3,
                     'duration (s)': np.concatenate((np.diff(matchedhe3), [run_stop])),
                     'offset (s)': matchedhe3-matchedli6}
            times['stop'] = times['start'] + times['duration (s)']
            times['supercycle'] = scycle

        ## get timestamps from sequencer
        elif mode in 'sequencer':

            # if sequencer is not enabled cause a stop transition
            df.inCycle *= df.sequencerEnabled

            # start counting only after first start flag
            df = df.loc[df.loc[df.cycleStarted > 0].index[0]:]

            # get start and end times
            df = df.diff()
            times = {'start': df.index[df.inCycle == 1],
                    'stop': df.index[df.inCycle == -1],
                    'supercycle': 0}

            # check lengths
            if len(times['start']) > len(times['stop']):
                times['stop'].append(df.index[-1])

        ## detector start times
        elif mode in 'he3':

            start = self.tfile[self.DET_NAMES['He3']['transitions']].cycleStartTime
            start = start.drop_duplicates()

            # setup output
            times = {'start': start,
                     'duration (s)': np.diff(np.concatenate((start, [run_stop])))
                    }
            times['stop'] = times['start'] + times['duration (s)']
            times['supercycle'] = self.tfile[self.DET_NAMES['He3']['transitions']].superCycleIndex

        ## detector start times
        elif mode in 'li6':

            start = self.tfile[self.DET_NAMES['Li6']['transitions']].cycleStartTime
            start = start.drop_duplicates()

            # setup output
            times = {'start': start,
                     'duration (s)': np.diff(np.concatenate((start, [run_stop]))),
                    }
            times['stop'] = times['start'] + times['duration (s)']
            times['supercycle'] = 0

        ## bad mode
        else:
            raise RuntimeError(f"Bad mode input {mode}")

        # convert to dataframe
        times = pd.DataFrame(times)
        times['duration (s)'] = times.stop - times.start
        times.index.name = 'cycle'

        # save
        self.cycle_param['cycle_times'] = times
        self.cycle_param['ncycles'] = len(times.index)

        # finish setting up cycle_param
        if 'period_end_times' not in self.cycle_param.keys():
            self.cycle_param['period_end_times'] = pd.DataFrame({0:times.loc[0, 'stop']}, index=[0])
            self.cycle_param['period_end_times'].index.name = 'period'
            self.cycle_param['period_end_times'].columns.name = 'cycle'

        if 'period_durations_s' not in self.cycle_param.keys():
            self.cycle_param['period_durations_s'] = pd.DataFrame({0:times.loc[0, 'duration (s)']}, index=[0])
            self.cycle_param['period_durations_s'].index.name = 'period'
            self.cycle_param['period_durations_s'].columns.name = 'cycle'

        return times



# experiment-general constants used in various calculations
# Derek Fujimoto
# Aug 2024

# The raw beam on and beam off times in EPICS are in units of beam buckets, which depend on the cyclotron frequency.  We kick a fraction of the beam buckets to the UCN target; there is a blanking period between buckets when there is no beam; we use that blanking period to ramp up/down the UCN kicker magnet.  So the fundamental unit of UCN beam kicking is beam buckets.
beam_bucket_duration_s = 0.00088801 # seconds

